from enum import Enum

import pydantic


class Type(str, Enum):
    """Attribute type."""

    STRING = "STRING"
    INTEGER = "INTEGER"
    TIMESTAMP = "TIMESTAMP"

    def __str__(self):
        return self.name


class ChangeType(str, Enum):
    """
    Attribute change type.
      + IGNORE: insert the latest data by business time, ignore updates.
      + UPDATE: insert the latest data by business time, update if business time.
      + NEW: save the full history of changes using SCD2.
    """

    IGNORE = "IGNORE"
    UPDATE = "UPDATE"
    NEW = "NEW"

    def __str__(self):
        return self.name


class Attribute(pydantic.BaseModel):
    """Attribute core representation."""

    name: str
    comment: str
    type: Type
    change_type: ChangeType
    group: str | None = None
    owner: str | None = None

    def __str__(self):
        return f"<{self.type} '{self.name}' change_type='{self.change_type}'>"

    def __hash__(self):
        return hash(f"{self.owner}.{self.name}.{self.type}")

    def __eq__(self, other):
        return hash(self) == hash(other)



import abc

from hnhm.core import Type, Attribute, ChangeType


class HnhmAttribute(abc.ABC):
    type: Type

    def __init__(
        self,
        *,
        comment: str,
        change_type: ChangeType,
        group: str | None = None,
    ):
        self.comment = comment
        self.change_type = change_type
        self.group = group

        self.owner = None
        self.name = None

    def to_core(self) -> Attribute:
        return Attribute(
            name=self.name,
            comment=self.comment,
            type=self.type,
            change_type=self.change_type,
            group=self.group,
        )

    def __set_name__(self, owner, name):
        self.owner = str(owner)
        self.name = name


class String(HnhmAttribute):
    type = Type.STRING


class Integer(HnhmAttribute):
    type = Type.INTEGER


class Timestamp(HnhmAttribute):
    type = Type.TIMESTAMP



from enum import Enum

import pydantic

from hnhm.core import (
    Sql,
    Entity,
    Storage,
    Mutation,
    HnhmError,
    CreateLink,
    LayoutType,
    RemoveLink,
    CreateGroup,
    RemoveGroup,
    CreateEntity,
    RemoveEntity,
    CreateAttribute,
    RemoveAttribute,
)

from .hnhm_link import HnhmLink
from .hnhm_entity import HnhmEntity


class PlanType(str, Enum):
    CREATE = "CREATE"
    REMOVE = "REMOVE"
    UPDATE = "UPDATE"


class PlanCollection(pydantic.BaseModel):
    type: PlanType
    mutations: list[Mutation]


class Plan(pydantic.BaseModel):
    entities_mutations: dict[str, PlanCollection]
    links_mutations: dict[str, PlanCollection]

    def is_empty(self):
        return not self.entities_mutations and not self.links_mutations

    @property
    def mutations_all(self):
        mutations: list[Mutation] = []
        for collection in self.entities_mutations.values():
            mutations.extend(collection.mutations)
        for collection in self.links_mutations.values():
            mutations.extend(collection.mutations)
        mutations = sorted(mutations, key=lambda m: m.priority)
        return mutations


class HnHm:
    def __init__(self, *, sql: Sql, storage: Storage):
        self.sql = sql
        self.storage = storage
        self.data = self.storage.load()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.storage.save(self.data)

    def plan(
        self,
        *,
        entities: list[HnhmEntity] | None = None,
        links: list[HnhmLink] | None = None,
    ) -> Plan:
        core_entities = {}
        for entity in entities or []:
            entity_core = entity.to_core()
            core_entities[entity_core.name] = entity_core

        core_links = {}
        for link in links or []:
            link_core = link.to_core()
            core_links[link_core.name] = link_core

        plan = Plan(entities_mutations={}, links_mutations={})

        # Entity: create if not exists
        for entity in core_entities.values():
            if entity.name in self.data.entities:
                continue

            # Create Entity
            mutations = [CreateEntity(entity=entity)]

            if entity.layout.type == LayoutType.HNHM:
                # Create Attribute
                for attribute in entity.attributes.values():
                    mutations.append(CreateAttribute(entity=entity, attribute=attribute))
                # Create Group
                for group in entity.groups.values():
                    mutations.append(CreateGroup(entity=entity, group=group))

            plan.entities_mutations[entity.name] = PlanCollection(
                type=PlanType.CREATE,
                mutations=mutations,
            )

        # Entity: create/remove Attribute/Group
        for entity in core_entities.values():
            if entity.name not in self.data.entities:
                continue

            if entity.layout.type != LayoutType.HNHM:
                continue

            attributes_state = self.data.entities[entity.name].attributes
            groups_state = self.data.entities[entity.name].groups

            mutations = []
            # Create Attribute
            for attribute_name, attribute in entity.attributes.items():
                if attribute_name not in attributes_state:
                    mutations.append(CreateAttribute(entity=entity, attribute=attribute))

            # Remove Attribute
            for attribute_name, attribute in attributes_state.items():
                if attribute_name not in entity.attributes:
                    mutations.append(RemoveAttribute(entity=entity, attribute=attribute))

            # Create Group
            for group_name, group in entity.groups.items():
                if group_name not in groups_state:
                    mutations.append(CreateGroup(entity=entity, group=group))

            # Remove Group
            for group_name, group in groups_state.items():
                if group_name not in entity.groups:
                    mutations.append(RemoveGroup(entity=entity, group=group))

            if mutations:
                plan.entities_mutations[entity.name] = PlanCollection(
                    type=PlanType.UPDATE,
                    mutations=mutations,
                )

        # Link: remove
        for link_name, link in self.data.links.items():
            if link_name not in core_links:
                plan.links_mutations[link_name] = PlanCollection(
                    type=PlanType.REMOVE,
                    mutations=[RemoveLink(link=link)],
                )

        # Entity: remove
        for entity_name, entity in self.data.entities.items():
            if entity_name not in core_entities:
                mutations = []

                if entity.layout.type == LayoutType.HNHM:
                    attributes_state = self.data.entities[entity_name].attributes
                    groups_state = self.data.entities[entity_name].groups
                    # Remove Attribute
                    for _, attribute_state in attributes_state.items():
                        mutations.append(
                            RemoveAttribute(entity=entity, attribute=attribute_state)
                        )
                    # Remove Group
                    for group_name, group in groups_state.items():
                        mutations.append(RemoveGroup(entity=entity, group=group))

                mutations.append(RemoveEntity(entity=entity))
                plan.entities_mutations[entity_name] = PlanCollection(
                    type=PlanType.REMOVE,
                    mutations=mutations,
                )

        # Link: create
        for link_name, link in core_links.items():
            if link_name not in self.data.links:
                plan.links_mutations[link.name] = PlanCollection(
                    type=PlanType.CREATE,
                    mutations=[CreateLink(link=link)],
                )

        return plan

    def apply(self, plan: Plan):
        for mutation in plan.mutations_all:
            sql = self.sql.generate_sql(mutation)

            match mutation:
                case CreateEntity(entity=entity):
                    assert entity.name not in self.data.entities
                    self.sql.execute(sql)
                    if entity.layout.type == LayoutType.HNHM:
                        attributes = {}
                        groups = {}
                    else:
                        attributes = entity.attributes
                        groups = entity.groups
                    self.data.entities[entity.name] = Entity(
                        name=entity.name,
                        layout=entity.layout,
                        doc=entity.doc,
                        keys=entity.keys,
                        attributes=attributes,
                        groups=groups,
                    )

                case RemoveEntity(entity=entity):
                    assert entity.name in self.data.entities
                    self.sql.execute(sql)
                    del self.data.entities[entity.name]

                case CreateAttribute(entity=entity, attribute=attribute):
                    assert entity.name in self.data.entities
                    assert (
                        attribute.name not in self.data.entities[entity.name].attributes
                    )
                    self.sql.execute(sql)
                    self.data.entities[entity.name].attributes[attribute.name] = attribute

                case RemoveAttribute(entity=entity, attribute=attribute):
                    assert entity.name in self.data.entities
                    assert attribute.name in self.data.entities[entity.name].attributes
                    self.sql.execute(sql)
                    del self.data.entities[entity.name].attributes[attribute.name]

                case CreateGroup(entity=entity, group=group):
                    assert entity.name in self.data.entities
                    assert group.name not in self.data.entities[entity.name].groups
                    self.sql.execute(sql)
                    self.data.entities[entity.name].groups[group.name] = group

                case RemoveGroup(entity=entity, group=group):
                    assert entity.name in self.data.entities
                    assert group.name in self.data.entities[entity.name].groups
                    self.sql.execute(sql)
                    del self.data.entities[entity.name].groups[group.name]

                case CreateLink(link=link):
                    assert link.name not in self.data.links
                    self.sql.execute(sql)
                    self.data.links[link.name] = link

                case RemoveLink(link=link):
                    assert link.name in self.data.links
                    self.sql.execute(sql)
                    del self.data.links[link.name]

                case _:
                    raise HnhmError(f"Unknown mutation: '{mutation}'")



import pydantic

from .link import Link
from .group import Group
from .entity import Entity
from .priority import Priority
from .attribute import Attribute


class Task(pydantic.BaseModel):
    """Flow task."""

    priority: Priority
    source: Entity
    business_time_field: Attribute
    keys_mapping: dict[Attribute, Attribute]


class LoadHub(Task):
    priority = Priority.FIRST
    target: Entity

    def __str__(self):
        return f"<LoadHub source='{self.source.name}' target='{self.target.name}'>"


class LoadAttribute(Task):
    priority = Priority.SECOND
    target: Entity
    source_attribute: Attribute
    target_attribute: Attribute

    def __str__(self):
        return (
            "<LoadAttribute"
            f" source='{self.source.name}'"
            f" target='{self.target.name}'"
            f" source_attribute='{self.source_attribute.name}'"
            f" target_attribute='{self.target_attribute.name}'>"
        )


class LoadGroup(Task):
    priority = Priority.SECOND
    target: Entity
    group: Group
    attributes_mapping: dict[Attribute, Attribute]

    def __str__(self):
        return (
            "<LoadGroup"
            f" source='{self.source.name}'"
            f" target='{self.target.name}'"
            f" group_name='{self.group.name}'>"
        )


class LoadLink(Task):
    priority = Priority.SECOND
    link: Link
    keys_mapping: dict[str, dict[Attribute, Attribute]]

    def __str__(self):
        return f"<LoadLink source='{self.source.name}' link_name='{self.link.name}'>"



SQL_TEMPLATE__LOAD_HUB = """
INSERT INTO hub__{{ target_name }}(
    {{ target_name }}_sk,
    {% for target_key in target_keys -%}
        {{ target_key }}_bk,
    {% endfor -%}
    valid_from
)
SELECT
    {{ source_sk }},
    {% for source_key in source_keys -%}
        {{ source_key }},
    {% endfor -%}
    {{ business_time_field }}
FROM
    stg__{{ source_name }}
ON CONFLICT
    ({{ target_name }}_sk) DO NOTHING
"""



SQL_TEMPLATE__CREATE_LINK = """
CREATE TABLE link__{{ name }}(
    {% for entity in entities -%}
        {{ entity }}_sk VARCHAR(32) NOT NULL,
    {% endfor -%}
    valid_from TIMESTAMPTZ NOT NULL,
    valid_to TIMESTAMPTZ,

    {% for entity in entities -%}
        CONSTRAINT fk_{{ entity }}_sk
            FOREIGN KEY({{ entity }}_sk) 
            REFERENCES hub__{{ entity }}({{ entity }}_sk)
            ON DELETE NO ACTION,
    {% endfor -%}

    PRIMARY KEY({{ primary_keys | join(', ') }})
)
"""



from .load_hub import SQL_TEMPLATE__LOAD_HUB
from .load_new import SQL_TEMPLATE__LOAD_NEW
from .create_hub import SQL_TEMPLATE__CREATE_HUB
from .create_link import SQL_TEMPLATE__CREATE_LINK
from .load_ignore import SQL_TEMPLATE__LOAD_IGNORE
from .load_update import SQL_TEMPLATE__LOAD_UPDATE
from .create_group import SQL_TEMPLATE__CREATE_GROUP
from .create_stage import SQL_TEMPLATE__CREATE_STAGE
from .create_attribute import SQL_TEMPLATE__CREATE_ATTRIBUTE



import sys
import importlib

import click

from .hnhm import Plan, PlanType
from .hnhm_registry import HnhmRegistry
from .core import (
    HnhmError,
    CreateLink,
    LayoutType,
    RemoveLink,
    CreateGroup,
    RemoveGroup,
    CreateEntity,
    RemoveEntity,
    CreateAttribute,
    RemoveAttribute,
)


def import_registry(module: str) -> HnhmRegistry:
    sys.path.append(".")
    imported_module = importlib.import_module(module)
    registry: HnhmRegistry = getattr(imported_module, "__registry__", None)
    if not registry:
        raise HnhmError(
            f"Failed to import registry from module: '{module}'."
            " Please, specify your registry via __registry__ object in your DWH module."
        )
    return registry


def print_plan(plan: Plan):
    if plan.is_empty():
        click.secho("Your DWH is up to date.", fg="green")
        return

    entities_mutations = sorted(plan.entities_mutations.items(), key=lambda kv: kv[0])
    links_mutations = sorted(plan.links_mutations.items(), key=lambda kv: kv[0])

    click.secho("Plan:")
    for entity_name, plan_collection in entities_mutations:
        if plan_collection.type == PlanType.CREATE:
            symbol, color = "+", "green"
        elif plan_collection.type == PlanType.REMOVE:
            symbol, color = "-", "red"
        elif plan_collection.type == PlanType.UPDATE:
            symbol, color = "[u]", "yellow"
        else:
            raise HnhmError()

        click.secho(f"\n{symbol} entity '{entity_name}'", fg=color)
        for entity_mutation in plan_collection.mutations:
            match entity_mutation:
                case CreateEntity(entity=entity):
                    if entity.layout.type == LayoutType.HNHM:
                        click.secho(f"  + hub '{entity.name}'", fg="green")
                    else:
                        click.secho(f"  + stage '{entity.name}'", fg="green")
                        for attribute in entity.attributes.values():
                            click.secho(f"    |attribute '{attribute.name}'", fg="green")

                case CreateAttribute(entity=_, attribute=attribute):
                    click.secho(f"  + attribute '{attribute.name}'", fg="green")

                case CreateGroup(entity=_, group=group):
                    click.secho(f"  + group '{group.name}'", fg="green")
                    for attribute in group.attributes.values():
                        click.secho(f"    |attribute '{attribute.name}'", fg="green")

                case RemoveEntity(entity=entity):
                    if entity.layout.type == LayoutType.HNHM:
                        click.secho(f"  - hub '{entity.name}'", fg="red")
                    else:
                        click.secho(f"  - stage '{entity.name}'", fg="red")
                        for attribute in entity.attributes.values():
                            click.secho(f"    |attribute '{attribute.name}'", fg="red")

                case RemoveAttribute(entity=_, attribute=attribute):
                    click.secho(f"  - attribute '{attribute.name}'", fg="red")

                case RemoveGroup(entity=_, group=group):
                    click.secho(f"  - group '{group.name}'", fg="red")
                    for attribute in group.attributes.values():
                        click.secho(f"    | attribute '{attribute.name}'", fg="red")

    for link_name, plan_collection in links_mutations:
        if plan_collection.type == PlanType.CREATE:
            symbol, color = "+", "green"
        elif plan_collection.type == PlanType.REMOVE:
            symbol, color = "-", "red"
        elif plan_collection.type == PlanType.UPDATE:
            symbol, color = "[u]", "yellow"
        else:
            raise HnhmError()

        click.secho(f"\n{symbol} link '{link_name}'", fg=color)
        for link_mutation in plan_collection.mutations:
            match link_mutation:
                case RemoveLink(link=link):
                    for element in link.elements:
                        click.secho(f"  |element '{element.entity.name}'", fg=color)
                case CreateLink(link=link):
                    for element in link.elements:
                        click.secho(f"  |element '{element.entity.name}'", fg=color)

    click.secho()


@click.group()
def cli():
    pass


@cli.command()
@click.argument("dwh_module")
def plan(dwh_module: str):
    registry = import_registry(dwh_module)

    with registry.hnhm as hnhm:
        plan = hnhm.plan(entities=registry.entities, links=registry.links)
        print_plan(plan)


@cli.command()
@click.argument("dwh_module")
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    default=False,
    show_default=True,
    help="Apply plan without confirmation.",
)
def apply(dwh_module: str, yes: bool):
    registry = import_registry(dwh_module)

    with registry.hnhm as hnhm:
        plan = hnhm.plan(entities=registry.entities, links=registry.links)
        print_plan(plan)

    if plan.is_empty():
        return

    if yes or click.confirm("Apply mutations?"):
        with registry.hnhm as hnhm:
            hnhm.apply(plan)

        click.secho("Applied!", fg="green")



SQL_TEMPLATE__LOAD_IGNORE = """
WITH data AS (
    SELECT
        {{ source_sk }} AS sk,
        {{ business_time_field }} AS valid_from,

        {% for source_attribute in source_attributes -%}
            {{ source_attribute }},
        {% endfor -%}

        ROW_NUMBER() OVER (
            PARTITION BY {{ source_sk }}
            ORDER BY     {{ business_time_field }} DESC
        ) AS row_number
    FROM
        {{ source_table }}
)
INSERT INTO {{ target_table }}(
    {{ target_sk }},
    {% for target_attribute in target_attributes -%}
    {{ target_attribute }},
    {% endfor -%}
    valid_from
)
SELECT
    d.sk,
    {% for source_attribute in source_attributes -%}
        d.{{ source_attribute }},
    {% endfor -%}
    d.valid_from
FROM
    data d
LEFT OUTER JOIN
    {{ target_table }} t
    ON d.sk = t.{{ target_sk }}
WHERE
    t.{{ target_sk }} IS NULL
    AND d.row_number = 1;
"""



from .hnhm import HnHm
from .hnhm_link import HnhmLink
from .hnhm_entity import HnhmEntity


class HnhmRegistry:
    def __init__(
        self,
        *,
        hnhm: HnHm,
        entities: list[HnhmEntity] | None = None,
        links: list[HnhmLink] | None = None,
    ):
        self.hnhm = hnhm
        self.entities = entities
        self.links = links

    def __str__(self):
        return "<HnhmRegistry>"



import pydantic

from .link import Link
from .group import Group
from .entity import Entity
from .priority import Priority
from .attribute import Attribute


class Mutation(pydantic.BaseModel):
    """Mutation represents DWH change."""

    priority: Priority


class CreateEntity(Mutation):
    priority = Priority.FIRST
    entity: Entity

    def __str__(self):
        return f"<CreateEntity '{self.entity.name}'>"


class RemoveEntity(Mutation):
    priority = Priority.SECOND
    entity: Entity

    def __str__(self):
        return f"<RemoveEntity '{self.entity.name}'>"


class CreateAttribute(Mutation):
    priority = Priority.SECOND
    entity: Entity
    attribute: Attribute

    def __str__(self):
        return f"<CreateAttribute '{self.attribute.name}' entity='{self.entity.name}'>"


class RemoveAttribute(Mutation):
    priority = Priority.FIRST
    entity: Entity
    attribute: Attribute

    def __str__(self):
        return f"<RemoveAttribute '{self.attribute.name}' entity='{self.entity.name}'>"


class CreateGroup(Mutation):
    priority = Priority.SECOND
    entity: Entity
    group: Group

    def __str__(self):
        return f"<CreateGroup '{self.group.name}' entity='{self.entity.name}'>"


class RemoveGroup(Mutation):
    priority = Priority.FIRST
    entity: Entity
    group: Group

    def __str__(self):
        return f"<RemoveGroup '{self.group.name}' entity='{self.entity.name}'>"


class CreateLink(Mutation):
    priority = Priority.SECOND
    link: Link

    def __str__(self):
        return f"<CreateLink '{self.link.name}'>"


class RemoveLink(Mutation):
    priority = Priority.FIRST
    link: Link

    def __str__(self):
        return f"<RemoveLink '{self.link.name}'>"



import abc

import pydantic

from .link import Link
from .entity import Entity


class HnhmStorageData(pydantic.BaseModel):
    """State data for hnhm."""

    entities: dict[str, Entity]
    links: dict[str, Link]


class Storage(abc.ABC):
    @abc.abstractmethod
    def load(self) -> HnhmStorageData:
        raise NotImplementedError

    @abc.abstractmethod
    def save(self, data: HnhmStorageData):
        raise NotImplementedError


class InMemStorage(Storage):
    def load(self) -> HnhmStorageData:
        return HnhmStorageData(entities={}, links={})

    def save(self, data: HnhmStorageData):
        pass



SQL_TEMPLATE__CREATE_HUB = """
CREATE TABLE hub__{{ name }}(
    {{ name }}_sk VARCHAR(32) PRIMARY KEY,
    {% for column, column_type in zip(columns, columns_types) -%}
        {{ column }}_bk {{ column_type }},
    {% endfor -%}
    valid_from TIMESTAMPTZ NOT NULL
)"""



from enum import Enum

import pydantic


class LayoutType(str, Enum):
    """
    Layout type.
      + STAGE: create single table with all provided attributes.
      + HNHM: create hub, attributes, and group tables.
    """

    STAGE = "STAGE"
    HNHM = "HNHM"

    def __str__(self):
        return self.name


class Layout(pydantic.BaseModel):
    """Layout core representation for entities and links."""

    name: str
    type: LayoutType | None = None

    def __str__(self):
        return f"<Layout '{self.name}' type='{self.type}'>"



import pydantic

from .attribute import Attribute, ChangeType


class Group(pydantic.BaseModel):
    """Group core representation."""

    name: str
    attributes: dict[str, Attribute]
    change_type: ChangeType

    def __str__(self):
        return f"<Group '{self.name}' change_type='{self.change_type}'>"



from .group import Group
from .entity import Entity
from .error import HnhmError
from .sql import Sql, FakeSql
from .link import Link, LinkElement
from .layout import Layout, LayoutType
from .attribute import Type, Attribute, ChangeType
from .storage import Storage, InMemStorage, HnhmStorageData
from .tasks import Task, LoadHub, LoadLink, LoadGroup, LoadAttribute
from .mutations import (
    Mutation,
    CreateLink,
    RemoveLink,
    CreateGroup,
    RemoveGroup,
    CreateEntity,
    RemoveEntity,
    CreateAttribute,
    RemoveAttribute,
)



SQL_TEMPLATE__CREATE_STAGE = """
CREATE TABLE stg__{{ name }}(
    {% for column, column_type in zip(columns, columns_types) -%}
        {{ column }} {{ column_type }} {% if not loop.last %} , {% endif %}
    {% endfor -%}
)"""



SQL_TEMPLATE__CREATE_ATTRIBUTE = """
CREATE TABLE attr__{{ entity_name }}__{{ attribute_name }}(
    {{ entity_name }}_sk VARCHAR(32) NOT NULL,
    {{ attribute_name }} {{ attribute_type }},
    {% for time_column in time_columns -%}
        {{ time_column }},
    {% endfor -%}
    CONSTRAINT fk_{{ entity_name }}_sk
      FOREIGN KEY({{ entity_name }}_sk) 
        REFERENCES hub__{{ entity_name }}({{ entity_name }}_sk)
        ON DELETE NO ACTION
)"""
